'''
Created on 2015/11/30

:author: hubo

Process events multi-threaded or multi-processed
'''
from vlcp.event.runnable import RoutineContainer
import multiprocessing
import threading
import socket
import signal
import errno
import sys
from vlcp.event.event import Event, withIndices
from vlcp.event.core import POLLING_IN, PollEvent
import functools
import traceback
if sys.version_info[0] >= 3:
    from queue import Full, Queue, Empty
else:
    from Queue import Full, Queue, Empty
import logging
from vlcp.event.connection import ResolveRequestEvent, ResolveResponseEvent

@withIndices('connector', 'type')
class ConnectorControlEvent(Event):
    ADDMATCHERS = 'addmatcher'
    REMOVEMATCHERS = 'removematcher'
    SETMATCHERS = 'setmatchers'
    STOPRECEIVE = 'stopreceive'
    STARTRECEIVE = 'startreceive'
import os

@withIndices()
class MoreResultEvent(Event):
    pass

class _Pipe(object):
    "Make a pipe looks like a socket"
    def __init__(self, fd, canread = True):
        self.canread = canread
        self.canwrite = not canread
        self.fd = fd
    def setblocking(self, blocking):
        import fcntl
        if blocking:
            fcntl.fcntl(self.fd, fcntl.F_SETFL, 0)
        else:
            fcntl.fcntl(self.fd, fcntl.F_SETFL, os.O_NONBLOCK)
    def fileno(self):
        return self.fd
    def send(self, data):
        try:
            os.write(self.fd, data)
        except OSError as exc:
            raise socket.error(*exc.args)
    def recv(self, size):
        try:
            return os.read(self.fd, size)
        except OSError as exc:
            raise socket.error(*exc.args)
    def close(self):
        if self.fd:
            os.close(self.fd)
            self.fd = None
    def __del__(self):
        self.close()

class Connector(RoutineContainer):
    def __init__(self, worker_start, matchers = (), scheduler = None, mp = True, inputlimit = 0, allowcontrol = True):
        '''
        :param worker_start: func(queuein, queueout), queuein is the input queue, queueout is the
               output queue. For queuein, each object is (event, matcher) tuple; For queueout, each
               object is a tuple of events to send. Every object in queuein must have a response in
               queueout.
        :param matcheres: match events to be processed by connector.
        :param scheduler: bind to specified scheduler
        :param mp: use multiprocessing if possible. For windows, multi-threading is always used.
        :param inputlimit: input queue size limit. 0 = infinite.
        :param allowcontrol: if True, the connector accepts ConnectorControlEvent for connector configuration.
        '''
        RoutineContainer.__init__(self, scheduler, True)
        self.worker_start = worker_start
        self.matchers = set(matchers)
        self.mp = mp
        self.inputlimit = inputlimit
        self.allowcontrol = allowcontrol
        self.stopreceive = False
        self.jobs = 0
        self._moreresult_matcher = MoreResultEvent.createMatcher()
    @staticmethod
    def wrap_signal(func):
        def f(*args, **kwargs):
            signal.signal(signal.SIGINT, signal.SIG_IGN)
            signal.signal(signal.SIGTERM, signal.SIG_DFL)
            return func(*args, **kwargs)
        return f
    @staticmethod
    def connector_pipe(queuein, pipeout, worker_start):
        try:
            queueout = multiprocessing.Queue()
            worker_start(queuein, queueout)
            while True:
                try:
                    events = queueout.get()
                    if events is None:
                        break
                    pipeout.send(events)
                except EOFError:
                    break
                except OSError as exc:
                    if exc.args[0] == errno.EINTR:
                        continue
                    else:
                        break
        finally:
            pipeout.close()
    @staticmethod
    def connector_socket(queuein, pipeout, worker_start):
        try:
            queueout = Queue()
            worker_start(queuein, queueout)
            while True:
                try:
                    events = queueout.get()
                    if events is None:
                        break
                    pipeout[0].put(events)
                    while True:
                        try:
                            events = queueout.get(False)
                            pipeout[0].put(events)
                        except Empty:
                            break
                    pipeout[1].send(b'\x00')
                except EOFError:
                    break
                except OSError as exc:
                    if exc.args[0] == errno.EINTR:
                        continue
                    else:
                        break
                except:
                    pass
        finally:
            pipeout[1].close()
    def enqueue(self, queue, event, matcher):
        queue.put((event, matcher), False)
        event.canignore = True
        self.jobs += 1
        if self.jobs == 1:
            self.scheduler.setPollingDaemon(self.pipein, False)
    def sendevents(self, events):
        moreResult = False
        for e in events:
            if self._moreresult_matcher.isMatch(e):
                moreResult = True
            else:
                self.scheduler.emergesend(e)
        if not moreResult:
            self.jobs -= 1
        if self.jobs == 0:
            self.scheduler.setPollingDaemon(self.pipein, True)
    def _createobjs(self, fork, mp):
        if mp:
            queue = multiprocessing.Queue(self.inputlimit)
        else:
            queue = Queue(self.inputlimit)
        if mp:
            pipein, pipeout = multiprocessing.Pipe()
            process = multiprocessing.Process(target=self.wrap_signal(self.connector_pipe), args=(queue, pipeout, self.worker_start))
            outqueue = None
        else:
            # Use a thread instead
            # Create a socket on localhost
            if fork:
                # Linux
                pifd, pofd = os.pipe()
                pipein = _Pipe(pifd, True)
                pipeout = _Pipe(pofd, False)
                pipein.setblocking(False)
                pipeout.setblocking(True)
            else:
                addrinfo = socket.getaddrinfo('localhost', 0, socket.AF_UNSPEC, socket.SOCK_STREAM, socket.IPPROTO_TCP, socket.AI_ADDRCONFIG|socket.AI_PASSIVE)
                socket_s = socket.socket(*addrinfo[0][0:2])
                socket_s.bind(addrinfo[0][4])
                socket_s.listen(1)
                addr_target = socket_s.getsockname()
                pipeout = socket.socket(*addrinfo[0][0:2])
                pipeout.setblocking(False)
                pipeout.connect_ex(addr_target)
                pipein, _ = socket_s.accept()
                pipein.setblocking(False)
                pipeout.setblocking(True)
                socket_s.close()
            outqueue = Queue()
            process = threading.Thread(target=self.connector_socket, args=(queue, (outqueue, pipeout), self.worker_start))
        process.daemon = True
        self.pipein = pipein
        return (process, queue, pipein, outqueue)
    def main(self):
        import os
        self.resolving = set()
        if hasattr(os, 'fork'):
            fork = True
        else:
            fork = False
        mp = self.mp and fork
        (process, queue, pipein, outqueue) = self._createobjs(fork, mp)
        try:
            process.start()
            self.scheduler.registerPolling(pipein, POLLING_IN, True)
            response_matcher = PollEvent.createMatcher(pipein.fileno(), PollEvent.READ_READY)
            error_matcher = PollEvent.createMatcher(pipein.fileno(), PollEvent.ERROR)
            control_matcher = ConnectorControlEvent.createMatcher(self)
            if self.allowcontrol:
                system_matchers = (response_matcher, error_matcher, control_matcher)
            else:
                system_matchers = (response_matcher, error_matcher)
            isFull = False
            isEOF = False
            while True:
                if not isEOF:
                    if isFull or self.stopreceive:
                        yield system_matchers
                    else:
                        yield tuple(self.matchers) + system_matchers
                    if self.matcher is error_matcher:
                        isEOF = True
                if isEOF:
                    self.scheduler.unregisterPolling(pipein, self.jobs == 0)
                    self.jobs = 0
                    pipein.close()
                    pipein = None
                    for m in self.waitWithTimeout(1):
                        yield m
                    if mp:
                        process.terminate()
                    (process, queue, pipein, outqueue) = self._createobjs(fork, mp)
                    process.start()
                    self.scheduler.registerPolling(pipein, POLLING_IN, True)
                    response_matcher = PollEvent.createMatcher(pipein.fileno(), PollEvent.READ_READY)
                    error_matcher = PollEvent.createMatcher(pipein.fileno(), PollEvent.ERROR)
                    if self.allowcontrol:
                        system_matchers = (response_matcher, error_matcher, control_matcher)
                    else:
                        system_matchers = (response_matcher, error_matcher)
                    isFull = False
                elif self.matcher is control_matcher:
                    if self.event.type == ConnectorControlEvent.ADDMATCHERS:
                        for m in self.event.matchers:
                            self.matchers.add(m)
                    elif self.event.type == ConnectorControlEvent.REMOVEMATCHERS:
                        for m in self.event.matchers:
                            self.matchers.discard(m)
                    elif self.event.type == ConnectorControlEvent.SETMATCHERS:
                        self.matchers = set(self.event.matchers)
                    elif self.event.type == ConnectorControlEvent.STOPRECEIVE:
                        self.stopreceive = True
                    else:
                        self.stopreceive = False
                elif self.matcher is response_matcher:
                    if mp:
                        while pipein.poll():
                            try:
                                events = pipein.recv()
                            except EOFError:
                                isEOF = True
                                break
                            self.sendevents(events)
                    else:
                        while True:
                            try:
                                if not pipein.recv(4096):
                                    isEOF = True
                                    break
                            except socket.error as exc:
                                if exc.errno == errno.EAGAIN or exc.errno == errno.EWOULDBLOCK:
                                    break
                                elif exc.errno == errno.EINTR:
                                    continue
                                else:
                                    isEOF = True
                                    break
                        while True:
                            try:
                                events = outqueue.get(False)
                            except Empty:
                                break
                            self.sendevents(events)
                    isFull = False
                else:
                    try:
                        self.enqueue(queue, self.event, self.matcher)
                    except Full:
                        isFull = True
        finally:
            if pipein is not None:
                self.scheduler.unregisterPolling(pipein, self.jobs == 0)
                pipein.close()
            if mp:
                process.terminate()
            else:
                queue.put(None)

class ThreadPool(object):
    def __init__(self, poolsize, worker, mp = False):
        self.poolsize = poolsize
        self.worker = worker
        self.mp = mp
    def create(self, queuein, queueout):
        import os
        if hasattr(os, 'fork') and self.mp:
            pool = [multiprocessing.Process(target=self.worker, args=(queuein, queueout)) for i in range(0, self.poolsize)]
        else:
            pool = [threading.Thread(target=self.worker, args=(queuein, queueout)) for i in range(0, self.poolsize)]
        for p in pool:
            p.daemon = True
            p.start()



def async_to_async(newthread = True, mp = False):
    def decorator(func):
        @functools.wraps(func)
        def handler(*args, **kwargs):
            if mp:
                p = multiprocessing.Process(target=func, args=args, kwargs=kwargs)
                p.daemon = True
                p.start()
            elif newthread:
                t = threading.Thread(target=func, args=args, kwargs=kwargs)
                t.daemon = True
                t.start()
            else:
                func(*args, **kwargs)
        return handler
    return decorator

def async_processor(func):
    @functools.wraps(func)
    def handler(queuein, queueout):
        while True:
            try:
                r = queuein.get(True)
                if r is None:
                    queueout.put(None)
                    break
                event, matcher = r
            except OSError as exc:
                if exc.args[0] == errno.EINTR:
                    continue
                else:
                    break
            try:
                func(event, matcher, queueout)
            except:
                # Ignore
                pass
    return handler

def processor_to_async(newthread = False, mp = False):
    def decorator(func):
        @functools.wraps(func)
        @async_to_async(newthread, mp)
        def handler(event, matcher, queueout):
            try:
                output = func(event, matcher)
            except:
                queueout.put(())
            else:
                queueout.put(output)
        return handler
    return decorator

def processor(func, newthread = False, mp = False):
    return async_processor(processor_to_async()(func))

def generator_to_async(newthread = True, mp = False):
    def decorator(func):
        @functools.wraps(func)
        @async_to_async(newthread, mp)
        def handler(event, matcher, queueout):
            for es in func(event, matcher):
                queueout.put(tuple(es) + (MoreResultEvent(),))
            queueout.put(())
        return handler
    return decorator
    
def async_generator_processor(func, newthread = True, mp = False):
    return async_processor(generator_to_async(func, newthread, mp))

@withIndices('pool')
class TaskEvent(Event):
    canignore = False

@withIndices('request')
class TaskDoneEvent(Event):
    pass

class TaskPool(Connector):
    "Thread pool for small tasks"
    @staticmethod
    def _generator_wrapper(func):
        def g(event, matcher):
            try:
                for es in func():
                    yield es
            except:
                (typ, val, tr) = sys.exc_info()
                yield (TaskDoneEvent(event, exception = val),)
            else:
                yield (TaskDoneEvent(event, result = None),)
        return g
    @staticmethod
    def _processor_wrapper(func):
        def f(event, matcher):
            try:
                return (TaskDoneEvent(event, result=func()),)
            except:
                (typ, val, tr) = sys.exc_info()
                return (TaskDoneEvent(event, exception = val),)
        return f
    @staticmethod
    def _async_wrapper(func):
        def f(event, matcher, queueout):
            def sender(es):
                queueout.put(tuple(es) + (MoreResultEvent(),))
            try:
                queueout.put((TaskDoneEvent(event, result=func(sender)),))
            except:
                (typ, val, tr) = sys.exc_info()
                queueout.put((TaskDoneEvent(event, exception = val),))
        return f
    @staticmethod
    @async_processor
    def _processor(event, matcher, queueout):
        if hasattr(event, 'task'):
            processor_to_async(getattr(event, 'newthread', False))(TaskPool._processor_wrapper(event.task))(event, matcher, queueout)
        elif hasattr(event, 'gen_task'):
            generator_to_async(getattr(event, 'newthread', True))(TaskPool._generator_wrapper(event.gen_task))(event, matcher, queueout)
        elif hasattr(event, 'async_task'):
            async_to_async(getattr(event, 'newthread', True))(TaskPool._async_wrapper(event.async_task))(event, matcher, queueout)
        else:
            queueout.put(())
    def __init__(self, scheduler=None, poolsize = 64):
        self.threadpool = ThreadPool(poolsize, self._processor, False)
        Connector.__init__(self, self.threadpool.create, matchers=(TaskEvent.createMatcher(self),), scheduler=scheduler,
                           mp=False, inputlimit=poolsize, allowcontrol=False)
    def runTask(self, container, task, newthread = False):
        "Run task() in task pool. Raise an exception or get return value in container.retvalue"
        e = TaskEvent(self, task=task, newthread = newthread)
        for m in container.waitForSend(e):
            yield m
        yield (TaskDoneEvent.createMatcher(e),)
        if hasattr(container.event, 'exception'):
            raise container.event.exception
        else:
            container.retvalue = container.event.result
    def runGenTask(self, container, gentask, newthread = True):
        "Run generator gentask() in task pool, yield customized events"
        e = TaskEvent(self, gen_task = gentask, newthread = newthread)
        for m in container.waitForSend(e):
            yield m
        yield (TaskDoneEvent.createMatcher(e),)
        if hasattr(container.event, 'exception'):
            raise container.event.exception
        container.retvalue = None
    def runAsyncTask(self, container, asynctask, newthread = True):
        "Run asynctask(sender) in task pool, call sender(events) to send customized events, and also return value in container.retvalue"
        e = TaskEvent(self, async_task = asynctask, newthread = newthread)
        for m in container.waitForSend(e):
            yield m
        yield (TaskDoneEvent.createMatcher(e),)
        if hasattr(container.event, 'exception'):
            raise container.event.exception
        else:
            container.retvalue = container.event.result

class Resolver(Connector):
    logger = logging.getLogger(__name__ + '.Resolver')
    def __init__(self, scheduler = None, poolsize = 256):
        rm = ResolveRequestEvent.createMatcher()
        Connector.__init__(self, ThreadPool(poolsize, Resolver.resolver).create, (rm,), scheduler, False, poolsize, False)
        self.resolving = set()
    @staticmethod
    @processor
    def resolver(event, matcher):
        params = event.request
        try:
            addrinfo = socket.getaddrinfo(*params)
            return (ResolveResponseEvent(params, response=addrinfo),)
        except:
            et, ev, tr = sys.exc_info()
            return (ResolveResponseEvent(params, error=ev),)
    def enqueue(self, queue, event, matcher):
        if event.request in self.resolving:
            # Duplicated resolves are finished in the same time
            event.canignore = True
        else:
            Connector.enqueue(self, queue, event, matcher)
            self.resolving.add(event.request)
    def sendevents(self, events):
        Connector.sendevents(self, events)
        for e in events:
            self.resolving.remove(e.request)



#env python
'''
Created on 2015/10/19

:author: hubo

Command-line entry 
'''

from __future__ import print_function
from vlcp.server import main

import sys
import os.path
# No argparse
import getopt

# Document
doc = '''Run VLCP server from command line
[python|pypy] vlcp.py [-f <configfile>] [-d] [-p <pidfile>] [-F <fork>] [startmodule] ...
[python|pypy] vlcp.py --help

Available options:
  -f            Configuration file position (default: /etc/vlcp.conf)
  -d            Start as a daemon (Need python-daemon support)
  -p            When start as a daemon, specify a pid file (default: /var/run/vlcp.pid, or configured in
                configuration file)
  -F            Fork to <fork> sub processes, may be used with protocol.default.reuseport option
  startmodule   Specify modules to be started, replace server.startup in configuration file
  -h,-?,--help  Show this help
'''
def usage():
    print(doc)
    sys.exit(2)
def parsearg():
    try:
        options, args = getopt.gnu_getopt(sys.argv[1:], 'f:p:F:?hd', 'help')
        configfile = None
        pidfile = '/var/run/vlcp.pid'
        daemon = False
        fork = None
        for k,v in options:
            if k == '--help' or k == '-?' or k == '-h':
                usage()
            elif k == '-f':
                configfile = v
            elif k == '-p':
                pidfile = v
            elif k == '-d':
                daemon = True
            elif k == '-F':
                fork = int(v)
        startup = None
        if args:
            startup = args
        return (configfile, daemon, pidfile, startup, fork)
    except getopt.GetoptError as exc:
        print(exc)
        usage()

def default_start():
    """
    Use `sys.argv` for starting parameters. This is the entry-point of `vlcp-start`
    """
    (config, daemon, pidfile, startup, fork) = parsearg()
    if config is None:
        if os.path.isfile('/etc/vlcp.conf'):
            config = '/etc/vlcp.conf'
        else:
            print('/etc/vlcp.conf is not found; start without configurations.')
    elif not config:
        config = None
    main(config, startup, daemon, pidfile, fork)


if __name__ == '__main__':
    default_start()



'''
Created on 2015/10/19

:author: hubo
'''
from vlcp.config import defaultconfig
from vlcp.protocol.http import Http, HttpRequestEvent
from vlcp.utils.http import statichttp
from vlcp.event.stream import MemoryStream
from vlcp.service.connection.tcpserver import TcpServerBase

@defaultconfig
class HttpServer(TcpServerBase):
    '''
    Create HTTP server on specified URLs, vHosts are supported.
    '''
    def __init__(self, server):
        TcpServerBase.__init__(self, server, Http)
        # Default Handlers
        @statichttp(self.apiroutine)
        def default404(env):
            for m in env.error(404, showerror = False):
                yield m
        @statichttp(self.apiroutine)
        def options(env):
            env.output(MemoryStream(b''))
            if False:
                yield
        def main():
            om = HttpRequestEvent.createMatcher(None, b'*', b'OPTIONS')
            dm = HttpRequestEvent.createMatcher()
            while True:
                yield (om, dm)
                if not self.apiroutine.event.canignore:
                    self.apiroutine.event.canignore = True
                    if self.apiroutine.matcher is om:
                        self.apiroutine.subroutine(options(self.apiroutine.event), False)
                    else:
                        self.apiroutine.subroutine(default404(self.apiroutine.event), False)
        self.apiroutine.main = main
        self.routines.append(self.apiroutine)
    


'''
Created on 2015/7/30

:author: hubo
'''

from namedstruct import *
from namedstruct.namedstruct import rawtype as _rawtype
import array
import socket

ethertype = enum('ethertype', globals(), uint16,
        ETHERTYPE_LOOP      = 0x0060,         #  Ethernet Loopback packet     
        ETHERTYPE_PUP       = 0x0200,         #  Xerox PUP packet             
        ETHERTYPE_PUPAT     = 0x0201,         #  Xerox PUP Addr Trans packet  
        ETHERTYPE_IP        = 0x0800,         #  Internet Protocol packet     
        ETHERTYPE_X25       = 0x0805,         #  CCITT X.25                   
        ETHERTYPE_ARP       = 0x0806,         #  Address Resolution packet    
        ETHERTYPE_BPQ       = 0x08FF,         #  G8BPQ AX.25 Ethernet Packet  [ NOT AN OFFICIALLY REGISTERED ID ] 
        ETHERTYPE_IEEEPUP   = 0x0a00,         #  Xerox IEEE802.3 PUP packet 
        ETHERTYPE_IEEEPUPAT = 0x0a01,         #  Xerox IEEE802.3 PUP Addr Trans packet 
        ETHERTYPE_DEC       = 0x6000,         #  DEC Assigned proto           
        ETHERTYPE_DNA_DL    = 0x6001,         #  DEC DNA Dump/Load            
        ETHERTYPE_DNA_RC    = 0x6002,         #  DEC DNA Remote Console       
        ETHERTYPE_DNA_RT    = 0x6003,         #  DEC DNA Routing              
        ETHERTYPE_LAT       = 0x6004,         #  DEC LAT                      
        ETHERTYPE_DIAG      = 0x6005,         #  DEC Diagnostics              
        ETHERTYPE_CUST      = 0x6006,         #  DEC Customer use             
        ETHERTYPE_SCA       = 0x6007,         #  DEC Systems Comms Arch       
        ETHERTYPE_TEB       = 0x6558,         #  Trans Ether Bridging         
        ETHERTYPE_RARP      = 0x8035,         #  Reverse Addr Res packet      
        ETHERTYPE_ATALK     = 0x809B,         #  Appletalk DDP                
        ETHERTYPE_AARP      = 0x80F3,         #  Appletalk AARP               
        ETHERTYPE_8021Q     = 0x8100,         #  802.1Q VLAN Extended Header  
        ETHERTYPE_IPX       = 0x8137,         #  IPX over DIX                 
        ETHERTYPE_IPV6      = 0x86DD,         #  IPv6 over bluebook           
        ETHERTYPE_PAUSE     = 0x8808,         #  IEEE Pause frames. See 802.3 31B 
        ETHERTYPE_SLOW      = 0x8809,         #  Slow Protocol. See 802.3ad 43B 
        ETHERTYPE_WCCP      = 0x883E,         #  Web-cache coordination protocol defined in draft-wilson-wrec-wccp-v2-00.txt 
        ETHERTYPE_PPP_DISC  = 0x8863,         #  PPPoE discovery messages     
        ETHERTYPE_PPP_SES   = 0x8864,         #  PPPoE session messages       
        ETHERTYPE_MPLS_UC   = 0x8847,         #  MPLS Unicast traffic         
        ETHERTYPE_MPLS_MC   = 0x8848,         #  MPLS Multicast traffic       
        ETHERTYPE_ATMMPOA   = 0x884c,         #  MultiProtocol Over ATM       
        ETHERTYPE_ATMFATE   = 0x8884,         #  Frame-based ATM Transport over Ethernet 
        ETHERTYPE_PAE       = 0x888E,         #  Port Access Entity (IEEE 802.1X) 
        ETHERTYPE_AOE       = 0x88A2,         #  ATA over Ethernet            
        ETHERTYPE_8021QS    = 0x88A8,         #  8021.Q Server VLAN Extended Header 
        ETHERTYPE_TIPC      = 0x88CA,         #  TIPC                         
        ETHERTYPE_1588      = 0x88F7,         #  IEEE 1588 Timesync 
        ETHERTYPE_FCOE      = 0x8906,         #  Fibre Channel over Ethernet  
        ETHERTYPE_TDLS      = 0x890D,         #  TDLS 
        ETHERTYPE_FIP       = 0x8914,         #  FCoE Initialization Protocol 
        ETHERTYPE_EDSA      = 0xDADA,         #  Ethertype DSA [ NOT AN OFFICIALLY REGISTERED ID ] 
        ETHERTYPE_AF_IUCV   = 0xFBFB          #  IBM af_iucv [ NOT AN OFFICIALLY REGISTERED ID ] 
)

arp_op_code = enum('arp_op_code', globals(), uint16,
        ARPOP_REQUEST   = 1,               #  ARP request.  
        ARPOP_REPLY     = 2,               #  ARP reply.  
        ARPOP_RREQUEST  = 3,               #  RARP request.  
        ARPOP_RREPLY    = 4,               #  RARP reply.  
        ARPOP_InREQUEST = 8,               #  InARP request.  
        ARPOP_InREPLY   = 9,               #  InARP reply.  
        ARPOP_NAK       = 10               #  (ATM)ARP NAK.  
)

ip_protocol = enum('ip_protocol', globals(), uint8, False,
    IPPROTO_IP = 0,             #  Dummy protocol for TCP               
    IPPROTO_ICMP = 1,           #  Internet Control Message Protocol    
    IPPROTO_IGMP = 2,           #  Internet Group Management Protocol   
    IPPROTO_IPIP = 4,           #  IPIP tunnels (older KA9Q tunnels use 94) 
    IPPROTO_TCP = 6,            #  Transmission Control Protocol        
    IPPROTO_EGP = 8,            #  Exterior Gateway Protocol            
    IPPROTO_PUP = 12,           #  PUP protocol                         
    IPPROTO_UDP = 17,           #  User Datagram Protocol               
    IPPROTO_IDP = 22,           #  XNS IDP protocol                     
    IPPROTO_DCCP = 33,          #  Datagram Congestion Control Protocol 
    IPPROTO_RSVP = 46,          #  RSVP protocol                        
    IPPROTO_GRE = 47,           #  Cisco GRE tunnels (rfc 1701,1702)    
    IPPROTO_IPV6   = 41,        #  IPv6-in-IPv4 tunnelling              
    IPPROTO_ESP = 50,          #  Encapsulation Security Payload protocol 
    IPPROTO_AH = 51,           #  Authentication Header protocol       
    IPPROTO_BEETPH = 94,       #  IP option pseudo header for BEET 
    IPPROTO_PIM    = 103,       #  Protocol Independent Multicast       
    IPPROTO_COMP   = 108,              #  Compression Header protocol 
    IPPROTO_SCTP   = 132,       #  Stream Control Transport Protocol    
    IPPROTO_UDPLITE = 136,      #  UDP-Lite (RFC 3828)                  
    IPPROTO_RAW    = 255        #  Raw IP packets                       
)


ETH_ALEN = 6

mac_addr = uint8[ETH_ALEN]

mac_addr.formatter = lambda x: ':'.join('%02x' % (n,) for n in x)
def _create_mac_addr(addr = None):
    if addr is None:
        return [0] * ETH_ALEN
    else:
        mac_array = [int(v,16) for v in addr.split(':')]
        return (mac_array + [0] * (ETH_ALEN - len(mac_array)))[:ETH_ALEN]

mac_addr.new = _create_mac_addr

mac_addr_bytes = prim(str(ETH_ALEN) + 's', 'mac_addr_bytes')

mac_addr_bytes.formatter = lambda x: ':'.join('%02x' % (c,) for c in bytearray(x.ljust(6, b'\x00')))
def _create_mac_addr_bytes(addr = None):
    if addr is None:
        return b'\x00' * ETH_ALEN
    else:
        return mac_addr.tobytes(mac_addr(addr))

mac_addr_bytes.new = _create_mac_addr_bytes

ip4_addr = prim('I', 'ip4_addr')

import socket as _socket

ip4_addr.formatter = lambda x: _socket.inet_ntoa(ip4_addr.tobytes(x))
def _ip4_addr_new(ip_addr = None):
    if ip_addr is None:
        return 0
    else:
        return ip4_addr.create(_socket.inet_aton(ip_addr))
ip4_addr.new = _ip4_addr_new

ip4_addr_bytes = prim('4s', 'ip4_addr_bytes')
ip4_addr_bytes.formatter = lambda x: _socket.inet_ntoa(x.ljust(4, b'\x00'))
def _ip4_addr_bytes_new(ip_addr = None):
    if ip_addr is None:
        return b'\x00\x00\x00\x00'
    else:
        return _socket.inet_aton(ip_addr)
ip4_addr_bytes.new = _ip4_addr_bytes_new

ip6_addr = uint8[16]
if hasattr(_socket, 'inet_ntop'):
    ip6_addr.formatter = lambda x: _socket.inet_ntop(_socket.AF_INET6, ip6_addr.tobytes(x.ljust(16, b'\x00')))
if hasattr(_socket, 'inet_pton'):
    def _ip6_addr_new(ip_addr = None):
        if ip_addr is None:
            return [0] * 16
        else:
            return ip6_addr.create(_socket.inet_pton(_socket.AF_INET6, ip_addr))
else:
    def _ip6_addr_new(ip_addr = None):
        if ip_addr is None:
            return [0] * 16
        else:
            fp, sep, lp = ip_addr.partition('::')
            if not sep:
                return [int(v,16) for v in fp.split(':')]
            else:
                fp_l = [int(v,16) for v in fp.split(':')]
                lp_l = [int(v,16) for v in lp.split(':')]
                if len(fp_l) + len(lp_l) > 8:
                    raise ValueError('Illegal IPv6 address')
                else:
                    return fp_l + [0] * (8 - len(fp_l) - len(lp_l)) + lp_l
ip6_addr.new = _ip6_addr_new
    
ip6_addr_bytes = prim('16s', 'ip6_addr')

if hasattr(_socket, 'inet_ntop'):
    ip6_addr.formatter = lambda x: _socket.inet_ntop(_socket.AF_INET6, x)
if hasattr(_socket, 'inet_pton'):
    def _ip6_addr_bytes_new(ip_addr = None):
        if ip_addr is None:
            return b'\x00' * 16
        else:
            return _socket.inet_pton(_socket.AF_INET6, ip_addr)
else:
    def _ip6_addr_bytes_new(ip_addr = None):
        if ip_addr is None:
            return b'\x00' * 16
        else:
            return ip6_addr.tobytes(_ip6_addr_new(ip_addr))
ip6_addr_bytes.new = _ip6_addr_bytes_new

ethernet_l2 = nstruct((mac_addr,'dl_dst'),
                         (mac_addr,'dl_src'),
                         (ethertype,'dl_type'),
                         name='ethernet_l2',
                         padding = 1,
                         size = lambda x: 18 if x.dl_type == ETHERTYPE_8021Q else
                                          22 if x.dl_type == ETHERTYPE_8021QS else
                                          14
                )

ethernet_l2_8021q = nstruct((uint16, 'vlan_tci'),
                                (ethertype, 'dl_type2'),
                                base = ethernet_l2,
                                name = "ethernet_l2_8021q",
                                criteria = lambda x: x.dl_type == ETHERTYPE_8021Q,
                                init = packvalue(ETHERTYPE_8021Q, 'dl_type')
                                )

ethernet_l2_8021ad = nstruct((uint16, 'svlan_tci'),
                             (ethertype, 'dl_type3'),
                             (uint16, 'vlan_tci'),
                             (ethertype, 'dl_type2'),
                             base = ethernet_l2,
                             name = 'ethernet_l2_8021ad',
                             criteria = lambda x: x.dl_type == ETHERTYPE_8021QS,
                             init = lambda x: (packvalue(ETHERTYPE_8021QS, 'dl_type')(x),
                                             packvalue(ETHERTYPE_8021Q, 'dl_type3')(x))
                             )

def dl_type(packet):
    if packet.dl_type == ETHERTYPE_8021Q:
        return packet.dl_type2
    else:
        return packet.dl_type

def vlan_vid(packet):
    if packet.dl_type == ETHERTYPE_8021Q:
        return (packet.vlan_tci & 0xfff) | 0x1000
    else:
        return 0

def vlan_pcp(packet):
    if packet.dl_type == ETHERTYPE_8021Q:
        return (packet.vlan_tci >> 13)
    else:
        return 0

def create_vlan_tci(vid, pcp = 0):
    return ((pcp & 0x7) << 13) | 0x1000 | (vid & 0xfff)


ethernet_l3 = nstruct((ethernet_l2,),
                      name = "ethernet_l3",
                      padding = 1,
                      classifier = lambda x: x.dl_type2 if x.dl_type == ETHERTYPE_8021Q or x.dl_type == ETHERTYPE_8021QS else x.dl_type)

arp_payload = nstruct(
        (uint16,'arp_hwtype'),
        (uint16,'arp_proto'),
        (uint8,'hw_len'),
        (uint8,'proto_len'),
        (arp_op_code,'arp_op'),
        (mac_addr,'arp_sha'),
        (ip4_addr,'arp_spa'),
        (mac_addr,'arp_tha'),
        (ip4_addr,'arp_tpa'),
        name = 'arp_payload',
        padding = 1,
        init = lambda x: (packvalue(1, 'arp_hwtype')(x),
                packvalue(ETHERTYPE_IP, 'arp_proto')(x),
                packvalue(6, 'hw_len')(x),
                packvalue(4, 'proto_len')(x))
        )

def create_packet(packettype, vid = None, pcp = 0, svid = None, spcp = 0, *args, **kwargs):
    if svid is not None:
        dl_type = packettype().dl_type
        return packettype((ethernet_l2, ethernet_l2_8021ad), *args,
                          dl_type2 = dl_type, vlan_tci = create_vlan_tci(vid, pcp),
                          svlan_tci = create_vlan_tci(svid, spcp),
                          **kwargs)
    elif vid is not None:
        dl_type = packettype().dl_type
        return packettype((ethernet_l2, ethernet_l2_8021q), *args,
                          dl_type2 = dl_type, vlan_tci = create_vlan_tci(vid, pcp),
                          **kwargs)
    else:
        return packettype(*args, **kwargs)
        
arp_packet = nstruct((arp_payload,),
                     base = ethernet_l3,
                     classifyby = (ETHERTYPE_ARP,),
                     name = 'arp_packet',
                     init = packvalue(ETHERTYPE_ARP, 'dl_type')
                     )

def checksum(data):
    if len(data) % 2:
        data += b'\x00'
    s = sum(array.array('H',data))
    s = (s & 0xffff) + (s >> 16)
    s += (s >> 16)
    return _socket.ntohs(~s & 0xffff)

def checksum2(*checksums):
    s = sum(checksums)
    s = (s & 0xffff) + (s >> 16)
    s += (s >> 16)
    return (s & 0xffff)

def ip4_checksum(x):
    data = x._tobytes(True)
    return checksum(data)

def _prepack_ip4_header(x):
    size = len(x)
    if size > 60:
        raise ValueError('IP header is too large')
    x.version_length = 0x40 | (size // 4)

ip_tos = enum('ip_tos', globals(), uint8, True,
                IPTOS_ECN_ECT1 = 0x01,
                IPTOS_ECN_ECT0 = 0x02,
                IPTOS_ECN_CE = 0x03,
                IPTOS_DSCP_AF11 =       0x28,
                IPTOS_DSCP_AF12 =       0x30,
                IPTOS_DSCP_AF13 =       0x38,
                IPTOS_DSCP_AF21 =       0x48,
                IPTOS_DSCP_AF22 =       0x50,
                IPTOS_DSCP_AF23 =       0x58,
                IPTOS_DSCP_AF31 =       0x68,
                IPTOS_DSCP_AF32 =       0x70,
                IPTOS_DSCP_AF33 =       0x78,
                IPTOS_DSCP_AF41 =       0x88,
                IPTOS_DSCP_AF42 =       0x90,
                IPTOS_DSCP_AF43 =       0x98,
                IPTOS_DSCP_EF   =       0xb8,
                IPTOS_CLASS_CS1 =       0x20,
                IPTOS_CLASS_CS2 =       0x40,
                IPTOS_CLASS_CS3 =       0x60,
                IPTOS_CLASS_CS4 =       0x80,
                IPTOS_CLASS_CS5 =       0xa0,
                IPTOS_CLASS_CS6 =       0xc0,
                IPTOS_CLASS_CS7 =       0xe0
              )

IPTOS_ECN_NOT_ECT = 0x00
IPTOS_ECN_MASK = 0x03
IPTOS_DSCP_MASK = 0xfc
IPTOS_CLASS_MASK = 0xe0
IPTOS_CLASS_CS0 = 0x00

ip_frag_offset = enum("ip_frag_offset", globals(), uint16, True,
                         IP_RF = 0x8000,
                         IP_DF = 0x4000,
                         IP_MF = 0x2000
                         )
IP_OFFMASK = 0x1fff

IP_FRAG_ANY   = (1 << 0) #  Is this a fragment? 
IP_FRAG_LATER = (1 << 1) #  Is this a fragment with nonzero offset? 

def ip_frag(packet):
    '''
    Not fragments:
        ip_frag(packet) == 0
        not ip_frag(packet)
        
    First packet of fragments:
        ip_frag(packet) == IP_FRAG_ANY
    
    Not first packet of fragments:
        ip_frag(packet) & IP_FRAG_LATER
    
    All fragments:
        ip_frag(packet) & IP_FRAG_ANY

    '''
    return ((packet.frag_off & IP_OFFMASK) and IP_FRAG_LATER) | ((packet.frag_off & (IP_OFFMASK | IP_MF)) and IP_FRAG_ANY)

ip4_header = nstruct(
        # 4bit version + 4bit length
        (uint8,'version_length'), 
        (ip_tos,'tos'),
        (uint16,'total_len'),
        (uint16,'identifier'),
        # 3bit frag + 13bit offset
        (ip_frag_offset,'frag_off'),
        (uint8,'ttl'),
        (ip_protocol,'proto'),
        (uint16,'checksum'),
        (ip4_addr,'ip_src'),
        (ip4_addr,'ip_dst'),
        (raw, 'options'),
        padding = 4,
        name = 'ip4_header',
        size = lambda x: (x.version_length & 0xf) * 4,
        prepack = _prepack_ip4_header,
        init = packvalue(0x45, 'version_length')
        )

def calcuate_ip4_checksum(x):
    x.checksum = 0
    x.checksum = ip4_checksum(x._get_embedded(ip4_header)) or 0xffff

def _prepack_ip4_payload(x):
    x.total_len = x._realsize()
    calcuate_ip4_checksum(x)

ip4_packet = nstruct((ip4_header,),
                     (raw, 'payload'),
                     base = ethernet_l3,
                     classifyby = (ETHERTYPE_IP,),
                     name = 'ip4_packet',
                     init = packvalue(ETHERTYPE_IP, 'dl_type'),
                     prepack = calcuate_ip4_checksum
                     )

# IP Packet without an Ethernet header
ip4_payload = nstruct((ip4_header,),
                      name = "ip4_payload",
                      padding = 1,
                      size = lambda x: x.total_len,
                      prepack = _prepack_ip4_payload
                      )

ip4_partial_payload = nstruct((ip4_header,),
                      name = "ip4_partial_payload",
                      padding = 1,
                      prepack = calcuate_ip4_checksum
                      )

ip4_payload_fragment = nstruct((raw, 'payload'),
                               base = ip4_payload,
                               criteria = lambda x: ip_frag(x) & IP_FRAG_LATER,
                               name = "ip4_payload_fragment"
                               )

ip4_partial_payload_fragment = nstruct((raw, 'payload'),
                               base = ip4_partial_payload,
                               criteria = lambda x: ip_frag(x) & IP_FRAG_LATER,
                               name = "ip4_partial_payload_fragment"
                               )

tcp_header_min = nstruct((uint16, 'sport'),
                          (uint16, 'dport'),
                          (uint32, 'seq'),
                          name = 'tcp_header_min',
                          padding = 1
                          )

tcp_flags = enum('tcp_flags', globals(), uint8, True,
                    TH_FIN        = 0x01,
                    TH_SYN        = 0x02,
                    TH_RST        = 0x04,
                    TH_PUSH       = 0x08,
                    TH_ACK        = 0x10,
                    TH_URG        = 0x20
                   )

def _pack_tcp_header_size(x):
    x.data_offset = ((len(x) // 4) << 4)

tcp_option_kind = enum('tcp_option_kind', globals(), uint8,
                       TCPOPT_EOL = 0,
                       TCPOPT_NOP = 1,
                       TCPOPT_MAXSEG = 2,
                       TCPOPT_WINDOW = 3,
                       TCPOPT_SACK_PERMITTED = 4,
                       TCPOPT_SACK = 5,
                       TCPOPT_TIMESTAMP = 8
                       )

_tcp_option_header = nstruct((tcp_option_kind, 'kind'),
                            name = "tcp_option_header",
                            padding = 1,
                            size = lambda x: 1 if (x.kind == TCPOPT_EOL or x.kind == TCPOPT_NOP) else 2
                            )

_tcp_option_header_length = nstruct((uint8, 'length'),
                                    name = "tcp_option_header_length",
                                    base = _tcp_option_header,
                                    criteria = lambda x: x.kind != TCPOPT_EOL and x.kind != TCPOPT_NOP
                                    )

def _prepack_tcp_option(x):
    s = x._realsize()
    if s > 1:
        x.length = s

tcp_option = nstruct((_tcp_option_header,),
                     name = "tcp_option",
                     padding = 1,
                     size = lambda x: 1 if (x.kind == TCPOPT_EOL or x.kind == TCPOPT_NOP) else x.length,
                     prepack = _prepack_tcp_option
                     )

def _init_tcp_option(kind):
    def _init_func(x):
        x._replace_embedded_type(_tcp_option_header, _tcp_option_header_length)
        x.kind = kind
    return _init_func

tcp_option_maxseg = nstruct((uint16, 'mss'),
                            name = 'tcp_option_maxseg',
                            base = tcp_option,
                            criteria = lambda x: x.kind == TCPOPT_MAXSEG,
                            init = _init_tcp_option(TCPOPT_MAXSEG)
                            )

tcp_option_window = nstruct((uint8, 'shift_cnt'),
                            name = 'tcp_option_window',
                            base = tcp_option,
                            criteria = lambda x: x.kind == TCPOPT_WINDOW,
                            init = _init_tcp_option(TCPOPT_WINDOW)
                            )

tcp_option_sack = nstruct((uint32[2][0], 'edges'),
                          name = 'tcp_option_sack',
                          base = tcp_option,
                          criteria = lambda x: x.kind == TCPOPT_SACK,
                          init = _init_tcp_option(TCPOPT_SACK)
                          )

tcp_option_timestamp = nstruct((uint32, 'tsval'),
                               (uint32, 'tsecr'),
                               name = 'tcp_option_timestamp',
                               base = tcp_option,
                               criteria = lambda x: x.kind == TCPOPT_TIMESTAMP,
                               init = _init_tcp_option(TCPOPT_TIMESTAMP))

def _format_tcp_options(x):
    # Parse tcp options
    options = tcp_option[0].create(x)
    result = []
    for o in options:
        od = dump(o, dumpextra = True, typeinfo = DUMPTYPE_NONE)
        anyvalue = False
        for k,v in od.items():
            if k not in ('kind', 'length') and not k.startswith('_'):
                result.append(k + ': ' + repr(v))
                anyvalue = True
        if '_extra' in od:
            result.append(str(od['kind']) + ': ' + repr(od['_extra']))
            anyvalue = True
        if not anyvalue and o.kind != TCPOPT_EOL and o.kind != TCPOPT_NOP:
            result.append(od['kind'])
    return ', '.join(result)

_tcp_option_raw = _rawtype()
_tcp_option_raw.formatter = _format_tcp_options

tcp_header = nstruct((tcp_header_min,),
                      (uint32, 'ack'),
                      (uint8, 'data_offset'),
                      (tcp_flags, 'tcp_flags'),
                      (uint16, 'tcp_win'),
                      (uint16, 'tcp_sum'),
                      (uint16, 'tcp_urp'),
                      (_tcp_option_raw, 'tcp_options'),
                      name = 'tcp_header',
                      padding = 4,
                      size = lambda x: (x.data_offset >> 4) * 4,
                      prepack = _pack_tcp_header_size
                      )

tcp_payload = nstruct((tcp_header,),
                      (raw, 'data'),
                      name = "tcp_payload",
                      padding = 1
                      )

ip4_pseudo_header = nstruct((ip4_addr, 'ip_src'),
                            (ip4_addr, 'ip_dst'),
                            (uint8,),
                            (ip_protocol, 'protocol'),
                            (uint16, 'length'),
                            name = "ip4_pseudo_header",
                            padding = 1
                            )

def tp4_checksum(payload, src, dst, protocol = IPPROTO_TCP):
    if hasattr(payload, '_tobytes'):
        data = payload._tobytes(True)
    else:
        data = payload
    return checksum2(checksum(ip4_pseudo_header(ip_src = src, ip_dst = dst, length = len(data), protocol = protocol)._tobytes()), checksum(data))

def _prepack_tcp4_payload(x):
    x.tcp_sum = 0
    x.tcp_sum = tp4_checksum(x, x.ip_src, x.ip_dst, IPPROTO_TCP) or 0xffff
    
_tcp_payload_ip4 = nstruct((tcp_payload,),
                           name = "_tcp_payload_ip4",
                           prepack = _prepack_tcp4_payload,
                           padding = 1,
                           lastextra = True
                           )

ip4_tcp_payload = nstruct((_tcp_payload_ip4,),
                            name = "ip4_tcp_payload",
                            base = ip4_payload,
                            criteria = lambda x: not ip_frag(x) and x.proto == IPPROTO_TCP,
                            lastextra = True,
                            init = packvalue(IPPROTO_TCP, 'proto'),
                          )

tcp_bestparse = nstruct((tcp_header_min,),
                       name = "tcp_bestparse",
                       padding = 1)

tcp_bestparse_min = nstruct((raw, 'otherpayload'),
                            base = tcp_bestparse,
                            name = "tcp_bestparse_min",
                            criteria = lambda x: x._realsize() < 20
                            )

tcp_bestparse_header = nstruct((uint32, 'ack'),
                      (uint8, 'data_offset'),
                      (tcp_flags, 'tcp_flags'),
                      (uint16, 'tcp_win'),
                      (uint16, 'tcp_sum'),
                      (uint16, 'tcp_urp'),
                      base = tcp_bestparse,
                      name = "tcp_bestparse_header",
                      criteria = lambda x: x._realsize() >= 20,
                      init = packvalue(5 << 4, 'data_offset')
                      )

tcp_bestparse_partial = nstruct((raw, 'otherpayload'),
                                base = tcp_bestparse_header,
                                name = "tcp_bestparse_partial",
                                criteria = lambda x: x._realsize() < (x.data_offset >> 4) * 4
                                )

_tcp_bestparse_full_options = nstruct((_tcp_option_raw, 'tcp_options'),
                                      name = "_tcp_bestparse_full_options",
                                      size = lambda x: max((x.data_offset >> 4) * 4 - 20, 0),
                                      padding = 4,
                                      prepack = packexpr(lambda x: ((len(x) // 4) << 4) + 5, 'data_offset')
                                      )

tcp_bestparse_full = nstruct((_tcp_bestparse_full_options,),
                             (raw, 'data'),
                                base = tcp_bestparse_header,
                                name = "tcp_bestparse_full",
                                criteria = lambda x: x._realsize() >= (x.data_offset >> 4) * 4
                                )

ip4_tcp_fragment_payload = nstruct((tcp_bestparse,),
                                   name = "ip4_tcp_fragment_payload",
                                   base = ip4_payload,
                                   criteria = lambda x: ip_frag(x) == IP_FRAG_ANY and x.proto == IPPROTO_TCP,
                                   init = packvalue(IPPROTO_TCP, 'proto'),
                                   lastextra = True
                                   )

ip4_tcp_partial_payload = nstruct((tcp_bestparse,),
                                   name = "ip4_tcp_partial_payload",
                                   base = ip4_partial_payload,
                                   criteria = lambda x: not (ip_frag(x) & IP_FRAG_LATER) and x.proto == IPPROTO_TCP,
                                   init = packvalue(IPPROTO_TCP, 'proto'),
                                   lastextra = True
                                   )

udp_header = nstruct((uint16, 'sport'),
                     (uint16, 'dport'),
                     (uint16, 'udp_len'),
                     (uint16, 'udp_sum'),
                     name = "udp_header",
                     padding = 1
                     )

udp_payload = nstruct((udp_header,),
                      (raw, 'data'),
                      name = "udp_payload",
                      padding = 1,
                      prepack = packrealsize('udp_len')
                      )

def _prepack_udp4_payload(x):
    x.udp_sum = 0
    x.udp_sum = tp4_checksum(x, x.ip_src, x.ip_dst, IPPROTO_UDP) or 0xffff

_udp_payload_ip4 = nstruct((udp_payload,),
                           name = "_udp_payload_ip4",
                           prepack = _prepack_udp4_payload,
                           padding = 1,
                           lastextra = True
                           )

ip4_udp_payload = nstruct((_udp_payload_ip4,),
                            name = "ip4_udp_payload",
                            base = ip4_payload,
                            criteria = lambda x: not ip_frag(x) and x.proto == IPPROTO_UDP,
                            lastextra = True,
                            init = packvalue(IPPROTO_UDP, 'proto')
                          )

ip4_udp_fragment_payload = nstruct((udp_header,),
                                   (raw, 'data'),
                                   name = "ip4_udp_fragment_payload",
                                   base = ip4_payload,
                                   criteria = lambda x: ip_frag(x) == IP_FRAG_ANY and x.proto == IPPROTO_UDP,
                                   init = packvalue(IPPROTO_UDP, 'proto')
                                   )

ip4_udp_partial_payload = nstruct((udp_header,),
                                   (raw, 'data'),
                                   name = "ip4_udp_partial_payload",
                                   base = ip4_partial_payload,
                                   criteria = lambda x: not (ip_frag(x) & IP_FRAG_LATER) and x.proto == IPPROTO_UDP,
                                   init = packvalue(IPPROTO_UDP, 'proto')
                                   )

icmp_type = enum('icmp_type', globals(), uint8,
                    ICMP_ECHOREPLY          = 0,       #  Echo Reply                   
                    ICMP_DEST_UNREACH       = 3,       #  Destination Unreachable      
                    ICMP_SOURCE_QUENCH      = 4,       #  Source Quench                
                    ICMP_REDIRECT           = 5,       #  Redirect (change route)      
                    ICMP_ECHO               = 8,       #  Echo Request                 
                    ICMP_TIME_EXCEEDED      = 11,      #  Time Exceeded                
                    ICMP_PARAMETERPROB      = 12,      #  Parameter Problem            
                    ICMP_TIMESTAMP          = 13,      #  Timestamp Request            
                    ICMP_TIMESTAMPREPLY     = 14,      #  Timestamp Reply              
                    ICMP_INFO_REQUEST       = 15,      #  Information Request          
                    ICMP_INFO_REPLY         = 16,      #  Information Reply            
                    ICMP_ADDRESS            = 17,      #  Address Mask Request         
                    ICMP_ADDRESSREPLY       = 18       #  Address Mask Reply           
                )

icmp_code_unreach = enum('icmp_code_unreach', globals(), uint8,
                    ICMP_NET_UNREACH        = 0,       #  Network Unreachable          
                    ICMP_HOST_UNREACH       = 1,       #  Host Unreachable             
                    ICMP_PROT_UNREACH       = 2,       #  Protocol Unreachable         
                    ICMP_PORT_UNREACH       = 3,       #  Port Unreachable             
                    ICMP_FRAG_NEEDED        = 4,       #  Fragmentation Needed/DF set  
                    ICMP_SR_FAILED          = 5,       #  Source Route failed          
                    ICMP_NET_UNKNOWN        = 6,
                    ICMP_HOST_UNKNOWN       = 7,
                    ICMP_HOST_ISOLATED      = 8,
                    ICMP_NET_ANO            = 9,
                    ICMP_HOST_ANO           = 10,
                    ICMP_NET_UNR_TOS        = 11,
                    ICMP_HOST_UNR_TOS       = 12,
                    ICMP_PKT_FILTERED       = 13,      #  Packet filtered 
                    ICMP_PREC_VIOLATION     = 14,      #  Precedence violation 
                    ICMP_PREC_CUTOFF        = 15       #  Precedence cut off 
                    )

icmp_code_redirect = enum('icmp_code_redirect', globals(), uint8,
                    ICMP_REDIR_NET          = 0,       #  Redirect Net                 
                    ICMP_REDIR_HOST         = 1,       #  Redirect Host                
                    ICMP_REDIR_NETTOS       = 2,       #  Redirect Net for TOS         
                    ICMP_REDIR_HOSTTOS      = 3        #  Redirect Host for TOS        
                    )

icmp_code_te = enum('icmp_code_te', globals(), uint8,
                    ICMP_EXC_TTL            = 0,       # TTL count exceeded
                    ICMP_EXC_FRAGTIME       = 1        # Fragment Reass time exceeded
                    )

icmp_header = nstruct((icmp_type, 'icmp_type'),
                      (uint8, 'icmp_code'),
                      (uint16, 'icmp_sum'),
                      name = "icmp_header",
                      padding = 1
                      )

icmp_bestparse = nstruct((icmp_header,),
                       name = "icmp_bestparse",
                       classifier = lambda x: x.icmp_type,
                       padding = 1
                       )

icmp_unreach = nstruct((uint32,),
                       (raw, 'data'),
                        classifyby = (ICMP_DEST_UNREACH,),
                        name = "icmp_unreach",
                        base = icmp_bestparse,
                        init = packvalue(ICMP_DEST_UNREACH, 'icmp_type'),
                        extend = {"icmp_code": icmp_code_unreach}
                    )

icmp_te = nstruct((uint32,),
                (raw, 'data'),
                classifyby = (ICMP_TIME_EXCEEDED,),
                name = "icmp_te",
                base = icmp_bestparse,
                init = packvalue(ICMP_TIME_EXCEEDED, 'icmp_type'),
                extend = {"icmp_code": icmp_code_te}
            )



icmp_parameter_prob = nstruct((uint8, 'icmp_pointer'),
                             (uint8[3],),
                             (raw, 'data'),
                             classifyby = (ICMP_PARAMETERPROB,),
                             name = 'icmp_parameter_prob',
                             base = icmp_bestparse,
                             init = packvalue(ICMP_PARAMETERPROB, 'icmp_type')
                        )

icmp_source_quench = nstruct((uint32,),
                      (raw, 'data'),
                      classifyby = (ICMP_SOURCE_QUENCH,),
                      name = 'icmp_source_quench',
                      base = icmp_bestparse,
                      init = packvalue(ICMP_SOURCE_QUENCH, 'icmp_type')
                      )

icmp_redirect = nstruct((uint32, 'gateway'),
                        (raw, 'data'),
                        classifyby = (ICMP_REDIRECT,),
                        name = 'icmp_redirect',
                        base = icmp_bestparse,
                        init = packvalue(ICMP_REDIRECT, 'icmp_type'),
                        extend = {"icmp_code": icmp_code_redirect}
                        )

icmp_echo = nstruct((uint16, 'icmp_id'),
                   (uint16, 'icmp_seq'),
                   (raw, 'data'),
                   classifyby = (ICMP_ECHO, ICMP_ECHOREPLY),
                   name = "icmp_echo",
                   base = icmp_bestparse,
                   init = packvalue(ICMP_ECHO, 'icmp_type')
                   )

icmp_timestamp = nstruct((uint16, 'icmp_id'),
                         (uint16, 'icmp_seq'),
                         classifyby = (ICMP_TIMESTAMP, ICMP_TIMESTAMPREPLY),
                         name = "icmp_timestamp",
                         base = icmp_bestparse,
                         init = packvalue(ICMP_TIMESTAMP, 'icmp_type')
                         )

icmp_timestamp_partial = nstruct((raw, 'otherpayload'),
                         name = "icmp_timestamp_partial",
                         base = icmp_timestamp,
                         criteria = lambda x: x._realsize() < 20
                         )

icmp_timestamp_full = nstruct((uint32, 'orig_timestamp'),
                              (uint32, 'recv_timestamp'),
                              (uint32, 'trans_timestamp'),
                              name = "icmp_timestamp_full",
                              base = icmp_timestamp,
                              criteria = lambda x: x._realsize() >= 20
                              )

icmp_info = nstruct((uint16, 'icmp_id'),
                     (uint16, 'icmp_seq'),
                     classifyby = (ICMP_INFO_REQUEST, ICMP_INFO_REPLY),
                     name = "icmp_info",
                     base = icmp_bestparse,
                     init = packvalue(ICMP_INFO_REQUEST, 'icmp_type')
                     )

def _prepack_icmp_payload(x):
    x.icmp_sum = 0
    x.icmp_sum = checksum(x._tobytes(True)) or 0xffff

_icmp_payload_ip4 = nstruct((icmp_bestparse,),
                           name = "_icmp_payload_ip4",
                           prepack = _prepack_icmp_payload,
                           padding = 1,
                           lastextra = True
                           )

ip4_icmp_payload = nstruct((_icmp_payload_ip4,),
                           name = "ip4_icmp_payload",
                            base = ip4_payload,
                            criteria = lambda x: not ip_frag(x) and x.proto == IPPROTO_ICMP,
                            lastextra = True,
                            init = packvalue(IPPROTO_ICMP, 'proto')
                          )

ip4_icmp_fragment_payload = nstruct((icmp_bestparse,),
                                    name = "ip4_icmp_fragment_payload",
                                   base = ip4_payload,
                                   criteria = lambda x: ip_frag(x) == IP_FRAG_ANY and x.proto == IPPROTO_ICMP,
                                   lastextra = True,
                                   init = packvalue(IPPROTO_ICMP, 'proto')
                                   )

ip4_icmp_partial_payload = nstruct((icmp_bestparse,),
                                   name = "ip4_icmp_partial_payload",
                                   base = ip4_partial_payload,
                                   criteria = lambda x: not (ip_frag(x) & IP_FRAG_LATER) and x.proto == IPPROTO_ICMP,
                                   lastextra = True,
                                   init = packvalue(IPPROTO_ICMP, 'proto')
                                   )

ethernet_l4 = nstruct((ethernet_l2,),
                      name = "ethernet_l4",
                      padding = 1,
                      classifier = lambda x: x.dl_type2 if x.dl_type == ETHERTYPE_8021Q or x.dl_type == ETHERTYPE_8021QS else x.dl_type)

ethernet_l7 = nstruct((ethernet_l2,),
                      name = "ethernet_l7",
                      padding = 1,
                      classifier = lambda x: x.dl_type2 if x.dl_type == ETHERTYPE_8021Q or x.dl_type == ETHERTYPE_8021QS else x.dl_type)

arp_packet_l4 = nstruct((arp_payload,),
                     base = ethernet_l4,
                     classifyby = (ETHERTYPE_ARP,),
                     name = 'arp_packet_l4',
                     init = packvalue(ETHERTYPE_ARP, 'dl_type'),
                     lastextra = True
                     )

arp_packet_l7 = nstruct((arp_payload,),
                     base = ethernet_l7,
                     classifyby = (ETHERTYPE_ARP,),
                     name = 'arp_packet_l7',
                     init = packvalue(ETHERTYPE_ARP, 'dl_type'),
                     lastextra = True
                     )

ip4_packet_l4 = nstruct((ip4_partial_payload,),
                     base = ethernet_l4,
                     classifyby = (ETHERTYPE_IP,),
                     name = 'ip4_packet_l4',
                     init = packvalue(ETHERTYPE_IP, 'dl_type'),
                     lastextra = True
                     )

ip4_packet_l7 = nstruct((ip4_payload,),
                     base = ethernet_l7,
                     classifyby = (ETHERTYPE_IP,),
                     name = 'ip4_packet_l7',
                     init = packvalue(ETHERTYPE_IP, 'dl_type'),
                     lastextra = True
                     )

def create_fragments_ip4(payload, mtu = 1500, options = '', fragment_options = '', proto = IPPROTO_UDP, *args, **kwargs):
    if hasattr(payload, '_tobytes'):
        payload = payload._tobytes()
    if len(payload) > 65536:
        raise ValueError('Payload is too large')
    ip4_header_size = len(ip4_header(options = options))
    ip4_header_frag_size = len(ip4_header(options = fragment_options))
    if len(payload) + ip4_header_size <= mtu:
        # No fragmentation
        return [ip4_payload_fragment(payload = payload,
                                    options = options,
                                    frag_off = 0,
                                    proto = proto,
                                    *args,
                                    **kwargs)]
    else:
        if mtu < ip4_header_size + 8 or mtu < ip4_header_frag_size + 8:
            raise ValueError('MTU is too small, at least %d is required, got %d' %
                             (max(ip4_header_size + 8, ip4_header_frag_size + 8), mtu))
        frags = []
        next_len = (mtu - ip4_header_size) // 8
        offset = 0
        totalsize = len(payload)
        frags.append(ip4_payload_fragment(payload = payload[offset * 8: (offset + next_len) * 8],
                                          options = options,
                                          frag_off = IP_MF,
                                          proto = proto,
                                          *args,
                                          **kwargs))
        while True:
            offset += next_len
            if totalsize - offset * 8 + ip4_header_frag_size <= mtu:
                frags.append(ip4_payload_fragment(payload = payload[offset * 8:],
                                                  options = fragment_options,
                                                  frag_off = offset,
                                                  proto = proto,
                                                  *args,
                                                  **kwargs))
                break
            else:
                next_len = (mtu - ip4_header_frag_size) // 8
                frags.append(ip4_payload_fragment(payload = payload[offset * 8: (offset + next_len) * 8],
                                                  options = fragment_options,
                                                  frag_off = IP_MF | offset,
                                                  proto = proto,
                                                  *args,
                                                  **kwargs))
        return frags

def create_fragments_ip4_packet(payload, mtu = 1500, options = '', fragment_options = '', proto = IPPROTO_UDP, **kwargs):
    if hasattr(payload, '_tobytes'):
        payload = payload._tobytes()
    if len(payload) > 65536:
        raise ValueError('Payload is too large')
    ip4_header_size = len(ip4_header(options = options))
    ip4_header_frag_size = len(ip4_header(options = fragment_options))
    if len(payload) + ip4_header_size <= mtu:
        # No fragmentation
        return [create_packet(ip4_packet, payload = payload,
                                    options = options,
                                    frag_off = 0,
                                    total_len = len(payload) + ip4_header_size,
                                    proto = proto,
                                    **kwargs)]
    else:
        if mtu < ip4_header_size + 8 or mtu < ip4_header_frag_size + 8:
            raise ValueError('MTU is too small, at least %d is required, got %d' %
                             (max(ip4_header_size + 8, ip4_header_frag_size + 8), mtu))
        frags = []
        next_len = (mtu - ip4_header_size) // 8
        offset = 0
        totalsize = len(payload)
        frags.append(create_packet(ip4_packet, payload = payload[offset * 8: (offset + next_len) * 8],
                                          options = options,
                                          frag_off = IP_MF,
                                          total_len = next_len * 8 + ip4_header_size,
                                          proto = proto,
                                          **kwargs))
        while True:
            offset += next_len
            if totalsize - offset * 8 + ip4_header_frag_size <= mtu:
                frags.append(create_packet(ip4_packet, payload = payload[offset * 8:],
                                                  options = fragment_options,
                                                  frag_off = offset,
                                                  total_len = totalsize - offset * 8 + ip4_header_frag_size,
                                                  proto = proto,
                                                  **kwargs))
                break
            else:
                next_len = (mtu - ip4_header_frag_size) // 8
                frags.append(create_packet(ip4_packet, payload = payload[offset * 8: (offset + next_len) * 8],
                                                  options = fragment_options,
                                                  frag_off = IP_MF | offset,
                                                  total_len = next_len * 8 + ip4_header_frag_size,
                                                  proto = proto,
                                                  **kwargs))
        return frags



'''
Created on 2016/6/22

:author: hubo
'''


from namedstruct import *
from vlcp.utils.ethernet import ip4_addr, ip4_addr_bytes, mac_addr_bytes
from vlcp.utils.netutils import parse_ip4_network

boot_op = enum('boot_op', globals(), uint8,
               BOOTREQUEST = 1,
               BOOTREPLY = 2)

BOOTP_MAGIC_COOKIE = 0x63825363

dhcp_tag = enum('dhcp_tag', globals(), uint8,
                OPTION_PAD = 0,
                OPTION_END = 255,
                OPTION_NETMASK = 1,
                OPTION_TIME_OFFSET = 2,
                OPTION_ROUTER = 3,
                OPTION_DNSSERVER = 6,
                OPTION_RLSERVER = 11,
                OPTION_HOSTNAME = 12,
                # Ignore many other options...
                OPTION_DOMAINNAME = 15,
                OPTION_MTU = 26,
                OPTION_BROADCAST = 28,
                OPTION_NTPSERVER = 42,
                OPTION_REQUESTED_IP = 50,
                OPTION_LEASE_TIME = 51,
                OPTION_OVERLOAD = 52,
                OPTION_MESSAGE_TYPE = 53,
                OPTION_SERVER_IDENTIFIER = 54,
                OPTION_REQUESTED_OPTIONS = 55,
                OPTION_MESSAGE = 56,
                OPTION_MAX_MESSAGE_SIZE = 57,
                OPTION_T1 = 58,
                OPTION_T2 = 59,
                OPTION_CLIENT_IDENTIFIER = 61,
                OPTION_CLASSLESSROUTE = 121
                )

dhcp_message_type = enum('dhcp_message_type', globals(), uint8,
                        DHCPDISCOVER           = 1,
                        DHCPOFFER              = 2,
                        DHCPREQUEST            = 3,
                        DHCPDECLINE            = 4,
                        DHCPACK                = 5,
                        DHCPNAK                = 6,
                        DHCPRELEASE            = 7,
                        DHCPINFORM             = 8
                        )

dhcp_overload = enum('dhcp_overload', globals(), uint8, True,
                     OVERLOAD_FILE = 1,
                     OVERLOAD_SNAME = 2)

class _EmptyOptionException(Exception):
    pass

def _no_empty(l):
    if not l:
        raise _EmptyOptionException()
    return l

def _prepack_dhcp_option(x):
    if x.tag != OPTION_PAD and x.tag != OPTION_END:
        x.length = 0
        x.length = x._realsize() - 2
        if x.length > 255:
            x.length = 255
    else:
        if hasattr(x, 'length'):
            delattr(x, 'length')

dhcp_option = nstruct((dhcp_tag, 'tag'),
                      (optional(uint8, 'length', lambda x: x.tag != OPTION_PAD and x.tag != OPTION_END),),
                      name = 'dhcp_option',
                      size = lambda x: getattr(x, 'length', -1) + 2,
                      prepack = _prepack_dhcp_option,
                      padding = 1
                      )

dhcp_option_single_byte = nstruct(name = 'dhcp_option_single_byte',
                                  base = dhcp_option,
                                  criteria = lambda x: x.tag in (OPTION_PAD, OPTION_END),
                                  init = packvalue(OPTION_PAD, 'tag'))

dhcp_option_address = nstruct((ip4_addr, 'value'),
                                 name = 'dhcp_option_address',
                                 base = dhcp_option,
                                 criteria = lambda x: x.tag in (OPTION_NETMASK, OPTION_REQUESTED_IP, OPTION_SERVER_IDENTIFIER, OPTION_BROADCAST),
                                 init = packvalue(OPTION_NETMASK, 'tag'))

dhcp_option_address._parse_from_value = lambda x: ip4_addr(x)

dhcp_time_value = enum('dhcp_time_value', globals(), uint32,
                       DHCPTIME_INFINITE = 0xffffffff
                  )

dhcp_option_time = nstruct((dhcp_time_value, 'value'),
                             name = 'dhcp_option_time',
                             base = dhcp_option,
                             criteria = lambda x: x.tag in (OPTION_TIME_OFFSET, OPTION_LEASE_TIME, OPTION_T1, OPTION_T2),
                             init = packvalue(OPTION_LEASE_TIME, 'tag')
                             )

dhcp_option_time._parse_from_value = lambda x: int(x) if x is not None and x != 'infinity' else DHCPTIME_INFINITE

dhcp_option_servers = nstruct((ip4_addr[0], 'value'),
                              name = 'dhcp_option_servers',
                              base = dhcp_option,
                              criteria = lambda x: OPTION_ROUTER <= x.tag <= OPTION_RLSERVER or x.tag in (OPTION_NTPSERVER,),
                              init = packvalue(OPTION_ROUTER, 'tag')
                              )

try:
    unicode
except:
    unicode = str

dhcp_option_servers._parse_from_value = lambda x: _no_empty([ip4_addr(x)] if isinstance(x, (str, unicode)) else [ip4_addr(v) for v in x])

dhcp_option_data = nstruct((raw, 'value'),
                           name = 'dhcp_option_name',
                           base = dhcp_option,
                           criteria = lambda x: x.tag in (OPTION_HOSTNAME, OPTION_DOMAINNAME, OPTION_MESSAGE, OPTION_CLIENT_IDENTIFIER),
                           init = packvalue(OPTION_HOSTNAME, 'tag'))


def _tobytes(s):
    if isinstance(s, bytes):
        return s
    elif isinstance(s, str):
        return s.decode('ascii')
    elif hasattr(s, '__iter__'):
        return create_binary(s, len(s))
    else:
        raise ValueError('Unsupported type: ')


dhcp_option_data._parse_from_value = _tobytes

dhcp_option_message_type = nstruct((dhcp_message_type, 'value'),
                                   name = 'dhcp_option_message_type',
                                   base = dhcp_option,
                                   criteria = lambda x: x.tag in (OPTION_MESSAGE_TYPE,),
                                   init = packvalue(OPTION_MESSAGE_TYPE, 'tag'))

dhcp_option_uint16 = nstruct((uint16, 'value'),
                          name = 'dhcp_option_uint16',
                          base = dhcp_option,
                          criteria = lambda x: x.tag in (OPTION_MTU, OPTION_MAX_MESSAGE_SIZE),
                          init = packvalue(OPTION_MTU, 'tag'))

dhcp_option_uint16._parse_from_value = lambda x: int(x)

dhcp_option_requested_options = nstruct((dhcp_tag[0], 'value'),
                                        name = 'dhcp_option_requested_options',
                                        base = dhcp_option,
                                        criteria = lambda x: x.tag == OPTION_REQUESTED_OPTIONS,
                                        init = packvalue(OPTION_REQUESTED_OPTIONS, 'tag'))

dhcp_option_overload = nstruct((dhcp_overload, 'value'),
                               name = 'dhcp_option_overload',
                               base = dhcp_option,
                               criteria = lambda x: x.tag == OPTION_OVERLOAD,
                               init = packvalue(OPTION_OVERLOAD, 'tag'))

def _prepack_destination(x):
    x.subnet = x.subnet[:(x.mask + 7) // 8]

def _destination_formatter(x):
    x['_cidr'] = ip4_addr_bytes.formatter(x.subnet) + '/' + str(x.mask)
    return x

_route_destination = nstruct((uint8, 'mask'),
                            (raw, 'subnet'),
                            name = '_route_destination',
                            padding = 1,
                            size = lambda x: (x.mask + 7) // 8 + 1,
                            prepack = _prepack_destination,
                            formatter = _destination_formatter
                            )

dhcp_route = nstruct((_route_destination,),
                     (ip4_addr, 'router'),
                     name = 'dhcp_route',
                     padding = 1
                     )

dhcp_classless_routes = nstruct((dhcp_route[0], 'value'),
                                name = 'dhcp_classless_routes',
                                base = dhcp_option,
                                criteria = lambda x: x.tag == OPTION_CLASSLESSROUTE,
                                init = packvalue(OPTION_CLASSLESSROUTE, 'tag'))

def _create_dhcp_route(cidr, router):
    network, mask = parse_ip4_network(cidr)
    return dhcp_route(subnet = ip4_addr.tobytes(network), mask = mask, router = ip4_addr(router))

dhcp_classless_routes._parse_from_value = lambda x: _no_empty([_create_dhcp_route(cidr, router) for cidr, router in x])

# According to RFC, options may be split into multiple parts
dhcp_option_partial = nstruct((dhcp_tag, 'tag'),
                      (optional(uint8, 'length', lambda x: x.tag != OPTION_PAD and x.tag != OPTION_END),),
                      (raw, 'data'),
                      name = 'dhcp_option_partial',
                      size = lambda x: getattr(x, 'length', -1) + 2,
                      prepack = _prepack_dhcp_option,
                      padding = 1
                      )

dhcp_flags = enum('dhcp_flags', globals(), uint16, True,
                  DHCPFLAG_BROADCAST = 1)

dhcp_payload = nstruct((boot_op, 'op'),
                       (uint8, 'htype'),
                       (uint8, 'hlen'),
                       (uint8, 'hops'),
                       (uint32, 'xid'),
                       (uint16, 'secs'),
                       (dhcp_flags, 'flags'),
                       (ip4_addr, 'ciaddr'),
                       (ip4_addr, 'yiaddr'),
                       (ip4_addr, 'siaddr'),
                       (ip4_addr, 'giaddr'),
                       (char[16], 'chaddr'),
                       (char[64], 'sname'),
                       (char[128], 'file'),
                       (uint32, 'magic_cookie'),    # 0x63825363
                       (dhcp_option_partial[0], 'options'),
                       name = 'dhcp_payload',
                       padding = 1,
                       extend = {'chaddr': mac_addr_bytes}
                       )

def reassemble_options(payload):
    '''
    Reassemble partial options to options, returns a list of dhcp_option
    '''
    options = []
    option_indices = {}
    def process_option_list(partials):
        for p in partials:
            if p.tag == OPTION_END:
                break
            if p.tag == OPTION_PAD:
                continue
            if p.tag in option_indices:
                # Reassemble the data
                options[option_indices[p.tag]][1].append(p.data)
            else:
                options.append((p.tag, [p.data]))
                option_indices[p.tag] = len(options) - 1
    # First process options field
    process_option_list(payload.options)
    if OPTION_OVERLOAD in option_indices:
        # There is an overload option
        data = b''.join(options[option_indices[OPTION_OVERLOAD]][1])
        overload_option = dhcp_overload.create(data)
        if overload_option & OVERLOAD_FILE:
            process_option_list(dhcp_option_partial[0].create(payload.file))
        if overload_option & OVERLOAD_SNAME:
            process_option_list(dhcp_option_partial[0].create(payload.sname))
    def _create_dhcp_option(tag, data):
        opt = dhcp_option(tag = tag)
        opt._setextra(data)
        opt._autosubclass()
        return opt
    return [_create_dhcp_option(tag, b''.join(data)) for tag,data in options]

def build_options(payload, options, maxsize = 576, overload = OVERLOAD_FILE | OVERLOAD_SNAME):
    '''
    Split a list of options 
    '''
    if maxsize < 576:
        maxsize = 576
    max_options_size = maxsize - 240
    options = [o for o in options if o.tag not in (OPTION_PAD, OPTION_END)]
    option_data = [(o.tag, o._tobytes()[2:]) for o in options]
    def split_options(option_data, limits):
        partial_options = []
        buffers = [0]
        if not options:
            return ([], True)
        def create_result():
            # Remove any unfinished partial options
            while partial_options and partial_options[-1][1]:
                partial_options.pop()
            buffers.append(len(partial_options))
            r = [[po for po,_ in partial_options[buffers[i]:buffers[i+1]]] for i in range(0, len(buffers) - 1)]
            while r and not r[-1]:
                r.pop()
            return r
        current_size = 0
        limit_iter = iter(limits)
        try:
            next_limit = next(limit_iter)
        except StopIteration:
            return ([], False)
        except GeneratorExit:
            return ([], False)
        for tag, data in option_data:
            data_size = 0
            nosplit = (len(data) <= 32)
            while True:
                next_size = min(next_limit - current_size - 2, 255, len(data) - data_size)
                if next_size < 0 or (next_size == 0 and data_size < len(data)) \
                        or (next_size < len(data) - data_size and nosplit):
                    try:
                        next_limit = next(limit_iter)
                    except StopIteration:
                        return (create_result(), False)
                    except GeneratorExit:
                        return (create_result(), False)
                    buffers.append(len(partial_options))
                    current_size = 0
                else:
                    partial_options.append((dhcp_option_partial(tag = tag, data = data[data_size : data_size + next_size]),
                                            (next_size < len(data) - data_size)))
                    data_size += next_size
                    current_size += next_size + 2
                    if data_size >= len(data):
                        break
        return (create_result(), True)
    # First try to fit all options in options field
    result, finished = split_options(option_data, [max_options_size - 1])
    if not finished:
        if overload & (OVERLOAD_FILE | OVERLOAD_SNAME):
            limits = [max_options_size - 4]
            if overload & OVERLOAD_FILE:
                limits.append(127)
            if overload & OVERLOAD_SNAME:
                limits.append(63)
            result, finished = split_options(option_data, limits)
    if not result:
        return
    elif len(result) <= 1:
        payload.options = result[0] + [dhcp_option_partial(tag = OPTION_END)]
    else:
        overload_option = 0
        if len(result) >= 2 and result[1]:
            overload_option |= OVERLOAD_FILE
            payload.file = dhcp_option_partial[0].tobytes(result[1] + [dhcp_option_partial(tag = OPTION_END)])
        if len(result) >= 3 and result[2]:
            overload_option |= OVERLOAD_SNAME
            payload.sname = dhcp_option_partial[0].tobytes(result[2] + [dhcp_option_partial(tag = OPTION_END)])
        payload.options = [dhcp_option_partial(tag = OPTION_OVERLOAD, data = dhcp_overload.tobytes(overload_option))] \
                        + result[0] + [dhcp_option_partial(tag = OPTION_END)]


def create_option_from_value(tag, value):
    dhcp_option.parser()
    fake_opt = dhcp_option(tag = tag)
    for c in dhcp_option.subclasses:
        if c.criteria(fake_opt):
            if hasattr(c, '_parse_from_value'):
                return c(tag = tag, value = c._parse_from_value(value))
            else:
                raise ValueError('Cannot set this DHCP option')
    else:
        fake_opt._setextra(_tobytes(value))
        return fake_opt

def create_dhcp_options(input_dict, ignoreError = False, generateNone = False):
    retdict = {}
    for k,v in dict(input_dict).items():
        try:
            if generateNone and v is None:
                retdict[k] = None
            else:
                try:
                    retdict[k] = create_option_from_value(k, v)
                except _EmptyOptionException:
                    if generateNone:
                        retdict[k] = None
        except Exception:
            if ignoreError:
                continue
            else:
                raise
    return retdict




'''
Created on 2016/2/23

:author: hubo
'''


from vlcp.config import defaultconfig
from vlcp.server.module import Module, api, depend, callAPI, ModuleNotification
from vlcp.event.runnable import RoutineContainer
from vlcp.service.sdn import ofpmanager
import vlcp.protocol.openflow.defs.openflow13 as of13
from vlcp.event.connection import ConnectionResetException
from vlcp.protocol.openflow.openflow import OpenflowProtocolException,\
    OpenflowAsyncMessageEvent
from vlcp.event.event import Event, withIndices
from contextlib import closing

def _bytes(s):
    return s.encode('ascii')

@withIndices('connection')
class OpenflowPortSynchronized(Event):
    pass

class OpenflowPortNotAppearException(Exception):
    pass

@defaultconfig
@depend(ofpmanager.OpenflowManager)
class OpenflowPortManager(Module):
    '''
    Manage Ports from Openflow Protocol
    '''
    service = True
    def __init__(self, server):
        Module.__init__(self, server)
        self.apiroutine = RoutineContainer(self.scheduler)
        self.apiroutine.main = self._manage_ports
        self.routines.append(self.apiroutine)
        self.managed_ports = {}
        self.createAPI(api(self.getports, self.apiroutine),
                       api(self.getallports, self.apiroutine),
                       api(self.getportbyno, self.apiroutine),
                       api(self.waitportbyno, self.apiroutine),
                       api(self.getportbyname, self.apiroutine),
                       api(self.waitportbyname, self.apiroutine),
                       api(self.resync, self.apiroutine)
                       )
        self._synchronized = False
    def _get_ports(self, connection, protocol, onup = False, update = True):
        ofdef = connection.openflowdef
        dpid = connection.openflow_datapathid
        vhost = connection.protocol.vhost
        add = []
        try:
            if hasattr(ofdef, 'ofp_multipart_request'):
                # Openflow 1.3, use ofp_multipart_request to get ports
                for m in protocol.querymultipart(ofdef.ofp_multipart_request(type=ofdef.OFPMP_PORT_DESC), connection, self.apiroutine):
                    yield m
                ports = self.managed_ports.setdefault((vhost, dpid), {})
                for msg in self.apiroutine.openflow_reply:
                    for p in msg.ports:
                        add.append(p)
                        ports[p.port_no] = p
            else:
                # Openflow 1.0, use features_request
                if onup:
                    # Use the features_reply on connection setup
                    reply = connection.openflow_featuresreply
                else:
                    request = ofdef.ofp_msg()
                    request.header.type = ofdef.OFPT_FEATURES_REQUEST
                    for m in protocol.querywithreply(request):
                        yield m
                    reply = self.apiroutine.retvalue
                ports = self.managed_ports.setdefault((vhost, dpid), {})
                for p in reply.ports:
                    add.append(p)
                    ports[p.port_no] = p
            if update:
                for m in self.apiroutine.waitForSend(OpenflowPortSynchronized(connection)):
                    yield m
                for m in self.apiroutine.waitForSend(ModuleNotification(self.getServiceName(), 'update',
                                                                         datapathid = connection.openflow_datapathid,
                                                                         connection = connection,
                                                                         vhost = protocol.vhost,
                                                                         add = add, remove = [],
                                                                         reason = 'connected')):
                    yield m
        except ConnectionResetException:
            pass
        except OpenflowProtocolException:
            pass
    def _get_existing_ports(self):
        for m in callAPI(self.apiroutine, 'openflowmanager', 'getallconnections', {'vhost':None}):
            yield m
        with closing(self.apiroutine.executeAll([self._get_ports(c, c.protocol, False, False) for c in self.apiroutine.retvalue if c.openflow_auxiliaryid == 0])) as g:
            for m in g:
                yield m
        self._synchronized = True
        for m in self.apiroutine.waitForSend(ModuleNotification(self.getServiceName(), 'synchronized')):
            yield m
    def _wait_for_sync(self):
        if not self._synchronized:
            yield (ModuleNotification.createMatcher(self.getServiceName(), 'synchronized'),)
    def _manage_ports(self):
        try:
            self.apiroutine.subroutine(self._get_existing_ports())
            conn_update = ModuleNotification.createMatcher('openflowmanager', 'update')
            port_status = OpenflowAsyncMessageEvent.createMatcher(of13.OFPT_PORT_STATUS, None, 0)
            while True:
                yield (conn_update, port_status)
                if self.apiroutine.matcher is port_status:
                    e = self.apiroutine.event
                    m = e.message
                    c = e.connection
                    if (c.protocol.vhost, c.openflow_datapathid) in self.managed_ports:
                        if m.reason == c.openflowdef.OFPPR_ADD:
                            # A new port is added
                            self.managed_ports[(c.protocol.vhost, c.openflow_datapathid)][m.desc.port_no] = m.desc
                            self.scheduler.emergesend(ModuleNotification(self.getServiceName(), 'update',
                                                                         datapathid = c.openflow_datapathid,
                                                                         connection = c,
                                                                         vhost = c.protocol.vhost,
                                                                         add = [m.desc], remove = [],
                                                                         reason = 'add'))
                        elif m.reason == c.openflowdef.OFPPR_DELETE:
                            try:
                                del self.managed_ports[(c.protocol.vhost, c.openflow_datapathid)][m.desc.port_no]
                                self.scheduler.emergesend(ModuleNotification(self.getServiceName(), 'update',
                                                                             datapathid = c.openflow_datapathid,
                                                                             connection = c,
                                                                             vhost = c.protocol.vhost,
                                                                             add = [], remove = [m.desc],
                                                                             reason = 'delete'))
                            except KeyError:
                                pass
                        elif m.reason == c.openflowdef.OFPPR_MODIFY:
                            try:
                                self.scheduler.emergesend(ModuleNotification(self.getServiceName(), 'modified',
                                                                             datapathid = c.openflow_datapathid,
                                                                             connection = c,
                                                                             vhost = c.protocol.vhost,
                                                                             old = self.managed_ports[(c.protocol.vhost, c.openflow_datapathid)][m.desc.port_no],
                                                                             new = m.desc,
                                                                             reason = 'modified'))
                            except KeyError:
                                self.scheduler.emergesend(ModuleNotification(self.getServiceName(), 'update',
                                                                             datapathid = c.openflow_datapathid,
                                                                             connection = c,
                                                                             vhost = c.protocol.vhost,
                                                                             add = [m.desc], remove = [],
                                                                             reason = 'add'))
                            self.managed_ports[(c.protocol.vhost, c.openflow_datapathid)][m.desc.port_no] = m.desc
                else:
                    e = self.apiroutine.event
                    for c in e.remove:
                        if c.openflow_auxiliaryid == 0 and (c.protocol.vhost, c.openflow_datapathid) in self.managed_ports:
                            self.scheduler.emergesend(ModuleNotification(self.getServiceName(), 'update',
                                                 datapathid = c.openflow_datapathid,
                                                 connection = c,
                                                 vhost = c.protocol.vhost,
                                                 add = [], remove = list(self.managed_ports[(c.protocol.vhost, c.openflow_datapathid)].values()),
                                                 reason = 'disconnected'))
                            del self.managed_ports[(c.protocol.vhost, c.openflow_datapathid)]
                    for c in e.add:
                        if c.openflow_auxiliaryid == 0:
                            self.apiroutine.subroutine(self._get_ports(c, c.protocol, True, True))
        finally:
            self.scheduler.emergesend(ModuleNotification(self.getServiceName(), 'unsynchronized'))
    def getports(self, datapathid, vhost = ''):
        "Return all ports of a specifed datapath"
        for m in self._wait_for_sync():
            yield m
        r = self.managed_ports.get((vhost, datapathid))
        if r is None:
            self.apiroutine.retvalue = None
        else:
            self.apiroutine.retvalue = list(r.values())
    def getallports(self, vhost = None):
        "Return all ``(datapathid, port, vhost)`` tuples, optionally filterd by vhost"
        for m in self._wait_for_sync():
            yield m
        if vhost is None:
            self.apiroutine.retvalue = [(dpid, p, vh) for (vh, dpid),v in self.managed_ports.items() for p in v.values()]
        else:
            self.apiroutine.retvalue = [(dpid, p, vh) for (vh, dpid),v in self.managed_ports.items() if vh == vhost for p in v.values()]
    def getportbyno(self, datapathid, portno, vhost = ''):
        "Return port with specified OpenFlow portno"
        for m in self._wait_for_sync():
            yield m
        self.apiroutine.retvalue = self._getportbyno(datapathid, portno, vhost)
    def _getportbyno(self, datapathid, portno, vhost = ''):
        ports = self.managed_ports.get((vhost, datapathid))
        if ports is None:
            return None
        else:
            return ports.get(portno)
    def waitportbyno(self, datapathid, portno, timeout = 30, vhost = ''):
        """
        Wait for the specified OpenFlow portno to appear, or until timeout.
        """
        for m in self._wait_for_sync():
            yield m
        def waitinner():
            ports = self.managed_ports.get((vhost, datapathid))
            if ports is None:
                for m in callAPI(self.apiroutine, 'openflowmanager', 'waitconnection', {'datapathid': datapathid, 'vhost':vhost, 'timeout': timeout}):
                    yield m
                c = self.apiroutine.retvalue
                ports = self.managed_ports.get((vhost, datapathid))
                if ports is None:
                    yield (OpenflowPortSynchronized.createMatcher(c),)
                ports = self.managed_ports.get((vhost, datapathid))
                if ports is None:
                    raise ConnectionResetException('Datapath %016x is not connected' % datapathid)
            if portno not in ports:
                yield (OpenflowAsyncMessageEvent.createMatcher(of13.OFPT_PORT_STATUS, datapathid, 0, _ismatch = lambda x: x.message.desc.port_no == portno),)
                self.apiroutine.retvalue = self.apiroutine.event.message.desc
            else:
                self.apiroutine.retvalue = ports[portno]
        for m in self.apiroutine.executeWithTimeout(timeout, waitinner()):
            yield m
        if self.apiroutine.timeout:
            raise OpenflowPortNotAppearException('Port %d does not appear on datapath %016x' % (portno, datapathid))
    def getportbyname(self, datapathid, name, vhost = ''):
        "Return port with specified port name"
        for m in self._wait_for_sync():
            yield m
        self.apiroutine.retvalue = self._getportbyname(datapathid, name, vhost)
    def _getportbyname(self, datapathid, name, vhost = ''):
        if not isinstance(name, bytes):
            name = _bytes(name)
        ports = self.managed_ports.get((vhost, datapathid))
        if ports is None:
            return None
        else:
            for p in ports.values():
                if p.name == name:
                    return p
            return None
    def waitportbyname(self, datapathid, name, timeout = 30, vhost = ''):
        """
        Wait for a port with the specified port name to appear, or until timeout
        """
        for m in self._wait_for_sync():
            yield m
        if not isinstance(name, bytes):
            name = _bytes(name)
        def waitinner():
            ports = self.managed_ports.get((vhost, datapathid))
            if ports is None:
                for m in callAPI(self.apiroutine, 'openflowmanager', 'waitconnection', {'datapathid': datapathid, 'vhost':vhost, 'timeout': timeout}):
                    yield m
                c = self.apiroutine.retvalue
                ports = self.managed_ports.get((vhost, datapathid))
                if ports is None:
                    yield (OpenflowPortSynchronized.createMatcher(c),)
                ports = self.managed_ports.get((vhost, datapathid))
                if ports is None:
                    raise ConnectionResetException('Datapath %016x is not connected' % datapathid)
            for p in ports.values():
                if p.name == name:
                    self.apiroutine.retvalue = p
                    return
            yield (OpenflowAsyncMessageEvent.createMatcher(of13.OFPT_PORT_STATUS, datapathid, 0, _ismatch = lambda x: x.message.desc.name == name),)
            self.apiroutine.retvalue = self.apiroutine.event.message.desc
        for m in self.apiroutine.executeWithTimeout(timeout, waitinner()):
            yield m
        if self.apiroutine.timeout:
            raise OpenflowPortNotAppearException('Port %r does not appear on datapath %016x' % (name, datapathid))
    def resync(self, datapathid, vhost = ''):
        '''
        Resync with current ports
        '''
        # Sometimes when the OpenFlow connection is very busy, PORT_STATUS message may be dropped.
        # We must deal with this and recover from it
        # Save current manged_ports
        if (vhost, datapathid) not in self.managed_ports:
            self.apiroutine.retvalue = None
            return
        else:
            last_ports = set(self.managed_ports[(vhost, datapathid)].keys())
        add = set()
        remove = set()
        ports = {}
        for _ in range(0, 10):
            for m in callAPI(self.apiroutine, 'openflowmanager', 'getconnection', {'datapathid': datapathid, 'vhost':vhost}):
                yield m
            c = self.apiroutine.retvalue
            if c is None:
                # Disconnected, will automatically resync when reconnected
                self.apiroutine.retvalue = None
                return
            ofdef = c.openflowdef
            protocol = c.protocol
            try:
                if hasattr(ofdef, 'ofp_multipart_request'):
                    # Openflow 1.3, use ofp_multipart_request to get ports
                    for m in protocol.querymultipart(ofdef.ofp_multipart_request(type=ofdef.OFPMP_PORT_DESC), c, self.apiroutine):
                        yield m
                    for msg in self.apiroutine.openflow_reply:
                        for p in msg.ports:
                            ports[p.port_no] = p
                else:
                    # Openflow 1.0, use features_request
                    request = ofdef.ofp_msg()
                    request.header.type = ofdef.OFPT_FEATURES_REQUEST
                    for m in protocol.querywithreply(request):
                        yield m
                    reply = self.apiroutine.retvalue
                    for p in reply.ports:
                        ports[p.port_no] = p
            except ConnectionResetException:
                break
            except OpenflowProtocolException:
                break
            else:
                if (vhost, datapathid) not in self.managed_ports:
                    self.apiroutine.retvalue = None
                    return
                current_ports = set(self.managed_ports[(vhost, datapathid)])
                # If a port is already removed
                remove.intersection_update(current_ports)
                # If a port is already added
                add.difference_update(current_ports)
                # If a port is not acquired, we do not add it
                acquired_keys = set(ports.keys())
                add.difference_update(acquired_keys)
                # Add and remove previous added/removed ports
                current_ports.difference_update(remove)
                current_ports.update(add)
                # If there are changed ports, the changed ports may or may not appear in the acquired port list
                # We only deal with following situations:
                # 1. If both lack ports, we add them
                # 2. If both have additional ports, we remote them
                to_add = acquired_keys.difference(current_ports.union(last_ports))
                to_remove = current_ports.intersection(last_ports).difference(acquired_keys)
                if not to_add and not to_remove and current_ports == last_ports:
                    break
                else:
                    add.update(to_add)
                    remove.update(to_remove)
                    current_ports.update(to_add)
                    current_ports.difference_update(to_remove)
                    last_ports = current_ports
        # Actual add and remove
        mports = self.managed_ports[(vhost, datapathid)]
        add_ports = []
        remove_ports = []
        for k in add:
            if k not in mports:
                add_ports.append(ports[k])
            mports[k] = ports[k]
        for k in remove:
            try:
                oldport = mports.pop(k)
            except KeyError:
                pass
            else:
                remove_ports.append(oldport)
        for m in self.apiroutine.waitForSend(ModuleNotification(self.getServiceName(), 'update',
                                                                 datapathid = datapathid,
                                                                 connection = c,
                                                                 vhost = vhost,
                                                                 add = add_ports, remove = remove_ports,
                                                                 reason = 'resync')):
            yield m
        self.apiroutine.retvalue = None
